"""
The shared callback logic

expects to be called as part of a class with the properties:
* instance.should_whitelist
* instance.category
* instance.analyzer
* instance.white_lister
"""
from qai.issues.meta import get_meta_value


def _is_valid(instance):
    assert instance.should_whitelist is not None
    assert instance.category is not None
    assert instance.analyzer is not None


def qallback(instance, segment, metadata, language):
    """
    Helper libs won't memoize this for us, because
    "Arguments to memoized function must be hashable"
    so we need to build it ourselves... weakass Python can't hash class instances
    """
    _is_valid(instance)

    if language not in instance.supported_languages:
        print(
            f"Received language {language}, supported Languages: {instance.supported_languages}"
        )
        return []

    if instance.should_whitelist:
        # Should we have whitelisting on?
        # this is a per-service setting;
        # e.g. Passive voice => no whitelisting
        # Gender bias => whitelisting
        enabled, meta_value, sub_groups = get_meta_value(instance.category, metadata)
        if enabled:
            # enabled/disabled is a per-project setting
            # e.g. customers can configer this in a GUI

            try:
                # allow new analyze method that has access to everything
                issues = instance.analyzer.analyze_with_metadata(
                    segment,
                    language=language,
                    meta_value=meta_value,
                    sub_groups=sub_groups,
                )
            except AttributeError:
                issues = instance.analyzer.analyze(segment)
            issues = instance.white_lister(issues, meta_value, sub_groups)
        else:
            # Value is None means this service is disabled
            # for this client
            issues = []
    else:
        # no whitelisting, just get issues
        issues = instance.analyzer.analyze(segment)

    return issues



from qai.qconnect.qallback import qallback


def parse_old_format(el):
    segment = el["data"]["content"]["segment"]
    metadata = el["data"]["meta"]
    language = el["data"]["content"]["languageCode"].split("-")[0]
    return segment, metadata, language


def parse_new_format(el):
    segment = el["content"]["segment"]
    metadata = el["meta"]
    language = el["content"]["languageCode"].split("-")[0]
    return segment, metadata, language


def add_issues_to_new_format(el, issues):
    el["issues"] = issues
    return el


def add_issues_to_old_format(el, issues):
    el["data"]["issues"] = issues
    return el


def add_issues_format_insensitive(instance, el):
    # We should pull some of this code out soon
    # Why write it then? because I don't trust the calling services
    # to do what Yakiv says will happen
    try:
        segment, metadata, language = parse_new_format(el)
        new_format = True
    except KeyError:
        print("could not parse in new format, trying old format")
        segment, metadata, language = parse_old_format(el)
        new_format = False
    issues = qallback(instance, segment, metadata, language)
    if new_format:
        el = add_issues_to_new_format(el, issues)
    else:
        el = add_issues_to_old_format(el, issues)
    return el



from typing import List, Optional, Tuple


def get_meta_value(
    category: str, meta: List[object]
) -> Tuple[bool, Optional[int], List]:
    """
    The business logic is:
        * We will get an array, meta
        * meta will either be the empty array, or have >= 1 element
            * if empty, the service is disabled
            * if nonempty, get meta[0]
            * meta[0].value is the usable value
            * CAN IGNORE meta[0].enabled - will never be zero, per Yakiv
    """
    enabled, value, sub_groups = False, None, []
    if not meta:
        pass
        print("empty meta array: assuming service is disabled")
    else:
        try:
            meta_el = meta[0]
        except IndexError as e:
            print("Yakiv swore this would not happen", e)
            return enabled, value, sub_groups
        try:
            styleGuideMeta = meta_el["styleGuideMeta"]
        except KeyError as e:
            print("Meta element has no styleGuideMeta prop", e)
            return enabled, value, sub_groups

        try:
            value = int(styleGuideMeta["value"])
            enabled = True
        except KeyError as e:
            value = None

        try:
            sub_groups = styleGuideMeta["subGroup"]
            enabled = True
        except KeyError as e:
            sub_groups = []

        if value == None and sub_groups == []:
            enabled = True
            print("Value and subGroups aren't set, assuming simple on/off service")

    return enabled, value, sub_groups




import uuid


def make_issue(word: str,
               start: int,
               end: int,
               issue_type: str,
               score,
               description='',
               suggestions=[],
               subCategory=''):
    '''
    reference:
    https://qordoba.atlassian.net/wiki/spaces/HOME/pages/840368141/Content-AI
    '''
    _uuid = str(uuid.uuid4())
    if not description:
        description = f'"{word}" is correlated with {issue_type} language.'
    issue = {
        "issueId": _uuid,
        "issueType": issue_type,
        "from": start,
        "until": end,
        "score": score,
        "suggestions": suggestions,
        "description": description,
    }
    
    if subCategory:
        issue["meta"] = { 
            "subCategory": subCategory 
            }

    return issue



import asyncio

import uvloop  # from requirements-dev.txt
from qai.issues.issues import make_issue
from qai.qconnect.qrest import QRest
from qai.qonfigs import get_configs
from qai.strings import get_strings_json


class MockAnalyzer:
    def __init__(self):
        pass

    def analyze(self, *args, **kwargs):
        return make_issue(
            "test",
            0,
            1,
            "test_issue_type",
            0,
            description="test issue description",
            suggestions=["X"],
        )


def whitelister(issues, meta_value, sub_groups=[]):
    if meta_value is not None:
        assert isinstance(meta_value, int)
    return issues


def test_server():
    server = QRest(
        MockAnalyzer(), category="test_qai", white_lister=whitelister
    ).get_future()

    loop = uvloop.new_event_loop()
    asyncio.set_event_loop(loop)
    _ = asyncio.ensure_future(server)

    loop.stop()  # I am pretty sure that this starts the loop for one loop



import math
import multiprocessing
import os
from pathlib import Path

from sanic import Sanic, response

from qai.funq import partial
from qai.version import __version__
from qai.issues.add_issues import add_issues_format_insensitive
from qai.qconnect.qremoteconnection import QRemoteConnection


def get_cpu_quota_within_docker():
    """
    By default, we use mp.cpu_count()
    HOWEVER, if there are cpu limits in certain paths, we assume those are
    from k8s/docker and override with those values
    """
    cpu_cores = multiprocessing.cpu_count()

    cfs_period = Path("/sys/fs/cgroup/cpu/cpu.cfs_period_us")
    cfs_quota = Path("/sys/fs/cgroup/cpu/cpu.cfs_quota_us")

    if cfs_period.exists() and cfs_quota.exists():
        # we are in a linux container with cpu quotas!
        with cfs_period.open("rb") as p, cfs_quota.open("rb") as q:
            p, q = int(p.read()), int(q.read())

            # get the cores allocated by dividing the quota
            # in microseconds by the period in microseconds
            cpu_cores = (
                math.ceil(q / p) if q > 0 and p > 0 else multiprocessing.cpu_count()
            )

    return cpu_cores


async def post_root(self, request):
    json_data = request.json
    try:
        # if it has items it is a dict
        _ = json_data.items()
        # if it is a dict, make it an array
        els = [json_data]
    except AttributeError:
        # can't call .items(), so it's already an array
        els = json_data
    resp_list = []
    for el in els:
        el = add_issues_format_insensitive(self, el)
        resp_list.append(el)
    return response.json(resp_list)


async def get_root(self, request):
    try:
        white_lister = self.white_lister
    except AttributeError:
        white_lister = None
    resp_obj = {
        "service": self.category,
        "status": "up",
        "host": self.host,
        "port": self.port,
        "qai_version": __version__,
        "white_lister": str(white_lister),
        "analyzer": str(self.analyzer),
    }
    return response.json(resp_obj)


class QRest(QRemoteConnection):
    def __init__(
        self,
        analyzer,
        category="",
        white_lister=None,
        host="0.0.0.0",
        port=5000,
        workers=get_cpu_quota_within_docker(),
        config_path=["conf", "config.json"],
    ):
        self.host = host
        self.port = port
        self.workers = workers

        config_file = os.path.join(os.getcwd(), *config_path)

        super().__init__(analyzer, category, white_lister, config_file)

        self.app = Sanic(__name__)

        # Add routes
        # This works, but it is better to make these after instantiating the class
        # that way you don't need partial application... see README
        unary_post_root = partial(post_root, self)
        unary_get_root = partial(get_root, self)
        self.app.add_route(unary_post_root, "/", methods=["POST"])
        self.app.add_route(unary_get_root, "/", methods=["GET"])

    def get_future(self):
        """
        Return a Future (Promise in JS land) that can be put on an event loop
        """
        return self.app.create_server(host=self.host, port=self.port)

    def connect(self):
        """
        Doesn't return, makes a blocking connection
        Only use if you are ONLY using REST
        This is a more robust REST server than get_future makes
        see:
        https://sanic.readthedocs.io/en/latest/sanic/deploying.html#asynchronous-support
        """
        if self.workers > 1:
            print(
                f"grpc, and hence AutoML, only support 1 worker, "
                + f"but you are using {self.workers}"
            )
        return self.app.run(host=self.host, port=self.port, workers=self.workers)



from typing import List, Optional, Tuple

from qai.issues.issues import make_issue

try:
    from spacy.tokens.span import Span
except ModuleNotFoundError as e:
    print(
        "\n\nQAI v2.0.0 and above assume spaCy is installed, but do not list it as a requirement"
        "If you want to use SpacyFactor, please install spaCy >=2.1.0\n\n"
    )
    raise e


class SpacyFactor(object):
    def __init__(self, issue_type: str, description: Optional[str] = None):
        self.issue_type = issue_type
        if description:
            self.description = description

    def __call__(self, spacy_span: Span, meta_subcategory=None, describer=None):
        """
        convenient wrapper around make_issue if you are using spaCy

        usage example:

        ```python
        from spacy.tokens import Span
        from app.factor import SpacyFactor


        SOV = SpacyFactor(
            "subject_object_verb_spacing",
            "Keep the subject, verb, and object of a sentence close together to help the reader understand the sentence."
        )

        Span.set_extension("score", default=0)
        Span.set_extension("suggestions", default=[])

        doc = nlp("Holders of the Class A and Class B-1 certificates will be entitled to receive on each Payment Date, to the extent monies are available therefor (but not more than the Class A Certificate Balance or Class B-1 Certificate Balance then outstanding), a distribution.")
        score = analyze(doc)
        if score is not None:
            span = Span(doc, 0, len(doc))  # or whichever TOKENS are the issue (don't have to worry about character indexes)
            span._.score = score
            span._.suggestions = get_suggestions(doc)
            issues = SOV(span)
        ```
        """
        text, start, end = spacy_span.text, spacy_span.start_char, spacy_span.end_char
        score = spacy_span._.score if spacy_span.has_extension("score") else 0
        suggestions = (
            spacy_span._.suggestions if spacy_span.has_extension("suggestions") else []
        )
        if describer:
            description = describer(spacy_span)
        else:
            description = self.description

        if meta_subcategory:
            subcategory = meta_subcategory
        else:
            subcategory = ""

        return make_issue(
            text,
            start,
            end,
            issue_type=self.issue_type,
            score=score,
            description=description,
            suggestions=suggestions,
            subCategory=subcategory,
        )
